/*
 *  Copyright (c) 2001 Dan Gudmundsson
 *  See the file "license.terms" for information on usage and redistribution
 *  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 *     $Id: esdl_driver.c,v 1.6 2003/01/13 16:12:38 bjorng Exp $
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <windows.h>
#include <io.h>
#else
#include <unistd.h>
#endif

#ifdef _OSX_COCOA
#include "CPS.h"
#include "SDLMain.h"
#include <Cocoa/Cocoa.h>
#include <objc/objc-runtime.h>
#endif

#include <errno.h>
#include <sys/types.h>
#include <fcntl.h>

#include <erl_driver.h>    /* Needed */
#include "esdl.h"

static ErlDrvData sdl_driver_start(ErlDrvPort port, char *buff);
static void sdl_driver_stop(ErlDrvData handle);
static void sdl_driver_output(ErlDrvData handle, char *buff, int bufflen);
static void sdl_driver_finish(void);
static int sdl_driver_control(ErlDrvData handle, unsigned int command, 
			      char* buf, int count, char** res, int res_size);

/*
** The driver struct
*/
ErlDrvEntry sdl_driver_driver_entry = {
    NULL,		   /* F_PTR init, N/A */
    sdl_driver_start,      /* L_PTR start, called when port is opened */
    sdl_driver_stop,       /* F_PTR stop, called when port is closed */
    sdl_driver_output,     /* F_PTR output, called when erlang has sent */
    NULL,                  /* F_PTR ready_input, called when input descriptor 
			      ready */
    NULL,                  /* F_PTR ready_output, called when output 
			      descriptor ready */
    "sdl_driver",          /* char *driver_name, the argument to open_port */
    sdl_driver_finish,     /* F_PTR finish, called when unloaded */
    NULL,                  /* void * that is not used (BC) */
    sdl_driver_control,    /* F_PTR control, port_control callback */
    NULL,                  /* F_PTR timeout, reserved */
    NULL,		   /* F_PTR outputv, reserved */
};

DRIVER_INIT(sdl_driver)
{
   return &sdl_driver_driver_entry;
}

ErlDrvData sdl_driver_start(ErlDrvPort port, char *buff)
{      
   sdl_data *data;   
   int i;

#ifdef _OSX_COCOA
   CPSProcessSerNum PSN;
   OSErr err;
#endif _OSX_COCOA

#ifdef _WIN32
   if ( SDL_RegisterApp("Erlang SDL", CS_BYTEALIGNCLIENT, 
			GetModuleHandle(NULL)) < 0 ) {
      fprintf(stderr, "WinMain() error %s \n", SDL_GetError());
      return(ERL_DRV_ERROR_GENERAL);
   }
#endif _WIN32
   
   data = malloc(sizeof(sdl_data));
   
   if(data == NULL) {
      fprintf(stderr, " Couldn't alloc mem\r\n");
      return(ERL_DRV_ERROR_GENERAL);  /* ENOMEM */      
   }
   else {
      set_port_control_flags(port, PORT_CONTROL_FLAG_BINARY);
      data->driver_data = port;
      data->fns   = init_fps();   
      
      data->noMemRefs = 0;
      for(i = 0; i < NOFFMEMREFS; i++)
	 data->memRefs[i] = NULL;
       
      data->op    = 0;
      data->len   = 0;
      data->buff  = NULL; 
       
#ifdef _OSX_COCOA
      // Removes the warnings about memory leakage
      // (not sure it actually works, though)
      data->release_pool = 
	objc_msgSend(objc_msgSend(objc_getClass("NSAutoreleasePool"), 
				  @selector(alloc)), @selector(init));
 
      data->app = objc_msgSend(objc_getClass("NSApplication"), 
			       @selector(sharedApplication));

      // To get a Menu & a dock icon : 
      err = CPSGetCurrentProcess(&PSN);
      assert(!(err = CPSSetProcessName(&PSN,"ESDL")));
      assert(!(err = CPSEnableForegroundOperation(&PSN,0x03,0x3C,0x2C,0x1103)));
      assert(!(err = CPSSetFrontProcess(&PSN)));

      // Makes the SDL window respond to keyboard events (???)
      //[NSAppleEventManager sharedAppleEventManager];

#endif _OSX_COCOA


      data->debug = 0;
   }
   return (ErlDrvData) data;
}

void sdl_driver_stop(ErlDrvData handle) 
{  
   int i;
   sdl_data *sd = ((sdl_data *)handle);
   SDL_Quit();
/*    if(sd->buff != NULL)  */
/*       Driverfree(sd->buff);    */
   free(sd->fns);
   for(i =0; i < NOFFMEMREFS; i++) {
      if(sd->memRefs[i] != NULL) {	
	 free(sd->memRefs[i]->ptr);
	 free(sd->memRefs[i]);
	 sd->memRefs[i] = NULL;
      }
   }

#ifdef _WIN32
   UnregisterClass("Erlang SDL", GetModuleHandle(NULL));
#endif _WIN32

   free(handle);
}

void sdl_driver_output(ErlDrvData handle, char *buff, int bufflen)
{
   do_command((sdl_data *)handle, 0, bufflen, buff);
}

void sdl_driver_finish(void) 
{
}

int sdl_driver_control(ErlDrvData handle, unsigned int command, 
			char* buf, int count, char** res, int res_size)
{
   sdl_data * sd = (sdl_data *) handle;
   sd->len = 0;    /* Release old data refs, should have been freed by erts */
   sd->buff = NULL;
   if(do_command(sd, command, count, buf) != 0) {
      (*res) = sd->buff;
      return sd->len;
   } else {  /* BUGBUG should probably clean up sd->buff */ 
      return -1;
   }
}

int sdl_send(sdl_data *sd, int len, char *buff)
{
   if(sd->buff == NULL) {
      fprintf(stderr, "ESDL INTERNAL ERROR: sdl_send in %s sent NULL buffer: %d\r\n",
	      sd->fns[sd->op]->str, len);
      return -1;
   }
   if(len > sd->len) 
      fprintf(stderr, "ESDL INTERNAL ERROR: sdl_send in %s allocated %d sent %d\r\n",
	      sd->fns[sd->op]->str, sd->len, len);
   /* Workaround that driver_control doesn't check lenght */
   ((ErlDrvBinary *) sd->buff)->orig_size = len;
   sd->len = len;

   return len;
}

char * sdl_getbuff(sdl_data *sd, int size, char *buff)
{  
   ErlDrvBinary * bin;
   sd->len = size;  
   bin = driver_alloc_binary(size); 
   sd->buff = bin;
   /* And return the pointer to the bytes */
   return bin->orig_bytes;
}
